// RELATED SOFTWARE: Scilab >= 5.5.0
// PACKAGE  : gui2bitmap : https://atoms.scilab.org/toolboxes/gui2bitmap
//            builds a png | jpg | gif | bmp image from a figure
//            possibly including uicontrols, with|without the window frame
// VERSION  : 1.2 - 2019-08-09
// AUTHOR   : Samuel GOUGEON
// COPYRIGHT: Samuel GOUGEON -  2019
// LICENSE  : CeCILL : http://www.cecill.info/licences/Licence_CeCILL_V2.1-en.txt
//
// DISCLAIMER: This toolbox is provided as is, with NO WARRANTY of any kind.
//             Use it to your own risks & responsability.
//

function gui2bitmap(varargin)
    // gui2bitmap(idFig, filename)
    // gui2bitmap(idFig, filename, withBorders)
    // gui2bitmap("setScale", scale)
    // gui2bitmap("setBorders", borders)
    // gui2bitmap("setMenu", parent)
    //
    // idFig    : Figure's handle. The figure must be not docked.
    //            On linux and MacOS, it is supposed to be not maximized.
    // filename : single string = pathanme of the image file to generate.
    //             The filename extension sets the image encoding.
    //             Supported output encodings: .png .jpg .gif .bmp
    //             Predefined constant paths SCI SCIHOME TMPDIR home..
    //             are supported.
    // withBorders: single boolean: %T to capture the whole figure with
    //              its borders. Default = %F
    // borders : boolean vector [withLeft, withBottom, withRight].
    //           default = [%T %T %T]
    // scale: single integer = percentage: default = 100
    // parent: graphical handle: get(0) => console, figure's handle,
    //         or another custom parent uimenu.

    sciVer = getversion("scilab");
    sciVer = evstr(strcat(msprintf("%d\n",sciVer(1:3)')));
    fname = "gui2bitmap"
    smode = getscilabmode() 
    if and(smode <> ["STD" "NW"]) then
        msg = _("%s can''t work in %s mode. Aborted.\n")
        error(msprintf(msg, fname, smode))
    end

    // PRIVATE FUNCTIONS
    // -----------------
    function r = is_maximized(idFig)
        // Windows only. Works wherever the taskbar is located
        fs = idFig.figure_size
        [xmax, ymax] = (getsystemmetrics("SM_CXMAXIMIZED"), ..
                        getsystemmetrics("SM_CYMAXIMIZED"));
        r = and(idFig.figure_size == [xmax ymax]);
    endfunction
    function yn = isdocked(fh)
        // For Scilab 5.5.2 (is included & named isDocked() in Scilab 6.0
        // http://fileexchange.scilab.org/toolboxes/360000
        // Shared internal function used by clf(), fchamp(), etc.
        //
        // fh: matrix of figures handles
        // yn: matrix of booleans
        //
        // yn(i) is %T if fh(i) is docked, %F otherwise
    
        sf = size(fh)
        fh = fh(:)
        fs = matrix(fh.figure_size,2,-1)'
        as = matrix(fh.axes_size,2,-1)'
    
        yn = (fh.dockable=="on" & ((fs(:,1)-as(:,1)) > 20)')
        // A test on vertical dimensions is more complicated, due to switchable
        // menubar, toolbar and infobar.
        yn = matrix(yn, sf)
    endfunction


    // CHECK INPUT ARGUMENTS
    // ---------------------
    rhs = argn(2);
    if rhs < 2 | rhs > 3  then
        msg = _("%s: Wrong number of input arguments: %d or %d expected.\n")
        error(msprintf(msg, fname,2,3))
    end
    // Argin #1
    // --------
    in = varargin(1);
    if and(type(in)<>[9 10])
        msg = _(fname, "%s: Argument #%d: Graphic handle or Text expected.\n")
        error(msprintf(msg, fname, 1))
    end
    in = in(1)
    if type(in)==9
        if in.type <> "Figure"
            msg = _("%s: Argument #%d: Graphic handle of type ''%s'' expected.\n")
            error(msprintf(msg, fname,1,"Figure"))
        end
        idFig = in
        if  isdocked(idFig) then
            msg = _(fname, "%s: The figure #%d is docked => Can''t proceed.\n")
            error(msprintf(msg, fname, idFig.figure_id))
        end
    else
        cmd = convstr(in)
        if and(convstr(cmd) <> ["setborders" "setscale" "setmenu"])
            msg = _(fname, "%s: Argument #%d:  ''%s'' expected.\n")
            error(msprintf(msg, fname,1,"Figure"))
        end
    end
    // Argin #2
    // --------
    in = varargin(2)
    if isdef("idFig","l") then
        // filename
        if type(in) <> 10 then
            msg = _(fname, "%s: Argument #%d: Text expected.\n")
            error(msprintf(msg, fname,2))
        end
        filename = pathconvert(in(1),%f) // support to SCI SCIHOME TMPDIR home..
        dname = fileparts(filename)
        if dname <> ""
            if ~isdir(dname) then
                mkdir(dname)
            end
        end
        // withBorders
        if rhs < 3 then
            withBorders = %f
        else
            in = varargin(3)
            if type(in) <> 4  then
                msg = _("%s: Argument #%d: Boolean expected.\n")
                error(msprintf(msg, fname, 3))
            end
            withBorders = in(1)
        end
        // Reading scale and borders prefs
        tmp = SCIHOME + filesep() + "gui2bitmap_preferences.xml"
        if isfile(tmp)
            attrib = ["scale" "borders"]
            if sciVer < 602 // 6.0.2
                doc = xmlRead(tmp);
                tmp = getPreferencesValue("/gui2bitmap", attrib, doc);
                xmlDelete(doc);
            else
                tmp = xmlGetValues("/gui2bitmap", attrib, tmp)
            end
            scale = evstr(tmp(1));
            bo = evstr(tmp(2));
        else
            scale = 100
            bo = [%T %T %T];
        end
        scale = scale/100;
    else
        if cmd=="setmenu"
            if getscilabmode()<>"STD"
                msg = _(fname, "%s: ""setMenu"": This syntax can be used only in STD Desktop execution mode.\n")
                error(msprintf(msg, fname))
            end
            if type(in)<>9
                msg = _("%s: Argument #%d: Graphic handle expected.\n")
                error(msprintf(msg, fname, 2))
            end
            // We could check here that the menu is not already set
            gui2bitmap_setmenu(in)
        elseif cmd=="setscale"
            if type(in)<>1
                msg = _("%s: Argument #%d: Decimal integer expected.\n")
                error(msprintf(msg, fname, 2))
            end
            in = in(1)
            if in < 1
                msg = _("%s: Argument #%d: Must be > %d.\n")
                error(msprintf(msg, fname, 2, 0))
            end
            // Updating (or setting) the preferences file
            gui2bitmap_setpref("scale", msprintf("%d", in))
        else // setBorders
            if type(in) <> 4
                msg = _("%s: Argument #%d: Boolean expected.\n")
                error(msprintf(msg, fname, 2))
            end
            if length(in) <> 3
                msg = _("%s: Argument #%d: Vector with %d elements expected.\n")
                error(msprintf(msg, fname, 2, 3))
            end
            // Updating (or setting) the preferences file
            gui2bitmap_setpref("borders", sci2exp(in))
        end
        return
    end


    // PROCESSING
    // ==========
    // Brings the GUI window on the front
    idFig.visible = "on";

    is_max = getos()=="Windows" // && is_maximized(idFig);
    if is_max, is_max = is_max & is_maximized(idFig); end // compatible Scilab 5
    fp = idFig.figure_position;
    fs = idFig.figure_size;
    as = idFig.axes_size;

    // Insure that the figure is not clipped out of screen on the left and top
    fp_old = fp;
    if getos()=="Windows" then
        screenW = getsystemmetrics("SM_CXMAXIMIZED");
        screenH = getsystemmetrics("SM_CYMAXIMIZED");
    else
        tmp = get(0,"screensize_px")
        [screenW, screenH] = (tmp(3), tmp(4));  // taskbar size not removed
    end

    drawlater
    if ~is_max then  // When maximized, the left and top borders may be at < 0
        if fp(1)<0 | (getos()=="Windows" & fs(1)<screenW & (fp(1)+fs(1))>screenW)
            idFig.figure_position(1) = 0;
            fp(1) = 0
            sleep(300)
        end
        if fp(2)<0 | (getos()=="Windows" & fs(2)<screenH & (fp(2)+fs(2))>screenH)
            idFig.figure_position(2) = 0;
            fp(2) = 0;
            sleep(300)
        end
    end

    // Compute the position and sizes of the area to grab
    bw = (fs(1)-as(1))/2;     // Borders Width. always correct
    header_height = fs(2) - as(2) - bw;
    // infobar_height = ?
    if ~withBorders then
        if idFig.infobar_visible=="off" then
            infobar_height = 0;
        else
            tmp1 = idFig.figure_size;
            idFig.infobar_visible = "off";
            tmp2 = idFig.figure_size;
            idFig.infobar_visible = "on";
            infobar_height = tmp1(2) - tmp2(2);
            sleep(500)
        end
        header_height = header_height - infobar_height
    end
    drawnow
    idFig.visible = "on";
    sleep(400)

    if withBorders then
        // Testée : OK (maximisée ou non, avec infobar ou non)
        [x, y, w, h] = (fp(1)+(1-bo(1))*bw, fp(2), ..
                        as(1)+(bo(1)+bo(3))*bw, fs(2)+(bo(2)-1)*bw);
        if is_max & bo(2)
            h = h - bw
            // left, right and bottom borders can't be grabbed
        end
    else
        [x, y, w, h] = (fp(1)+bw, fp(2)+header_height, as(1), as(2));
    end
    [x, y, w, h] = (x*scale, y*scale, w*scale, h*scale);

    // Grabs the figure into an image
    jimport java.awt.Rectangle;
    jimport java.awt.Robot;
    jimport java.awt.image.BufferedImage;
    jimport java.io.File;
    jimport javax.imageio.ImageIO;
    bounds = Rectangle.new(x, y, w, h);
    imtype = strsubst(fileparts(filename,"extension"),".","");
    rob = jnewInstance(Robot);
    ImageIO.write(rob.createScreenCapture(bounds), imtype, File.new(filename));

    // Finish
    jremove bounds rob BufferedImage File ImageIO Rectangle Robot
    idFig.figure_position = fp_old;
endfunction

// --------------------------------------------------

function gui2bitmap_setmenu(idFig)
    console = get(0);
    oldshh = get(console, "showhiddenhandles");
    set(console, "showhiddenhandles","on");
    if idFig.type=="Figure" then
        parent = idFig.children($)
    else    // console, etc
        parent =  idFig;
    end
    if idFig.type=="Figure"
        if parent.type<>"uimenu"
            msg = _("gui2bitmap", "%s: The figure #%d is already plotted.\n\tThe menu must be set BEFORE any plotting.\n");
            warning(msprintf(msg, "gui2bitmap", idFig.figure_id))
            set(console, "showhiddenhandles", oldshh)
            return
        else
            idFig.menubar_visible = "on";
        end
    end
    if idFig.type=="Console" then
        addmenu("gui2bitmap")
        gui2bitmapMenu = idFig.children(1)
        label = [_("gui2bitmap", "Export gcf() with borders") ..
                 _("gui2bitmap", "Export gcf() without borders")];
        target = "h = gcf(); h.visible=""on""; if length(winsid())>1, twinkle(h,2), end; "
    else
        gui2bitmapMenu = uimenu(parent, "Label","gui2bitmap");
        label = [_("gui2bitmap", "Export with borders") ..
                 _("gui2bitmap", "Export without borders")];
        target = "h = gcbo.parent.parent.parent; "
    end
    cb = target + "fname = uiputfile(""*.png"", pwd());" + ..
         "if fname<>"""", gui2bitmap(h, fname %s); end";
    uim1 = uimenu(gui2bitmapMenu, "Label",label(1), "Callback",msprintf(cb,", %T"));
    uim2 = uimenu(gui2bitmapMenu, "Label",label(2), "Callback",msprintf(cb,""));
    // Now move the item after the "Export" one: http://bugzilla.scilab.org/16169
    //for i = 5:-1:2
    //    swap_handles(idFig.children($).children(i), idFig.children($).children(1))
    //end
    //
    set(console, "showhiddenhandles", oldshh)
endfunction

// --------------------------------------------------

function gui2bitmap_setpref(attribute, value)
    tmp = SCIHOME + filesep() + "gui2bitmap_preferences.xml"
    //<?xml version="1.0" encoding="UTF-8"?>
    //    <gui2bitmap scale="100" borders="[%T %T %T]"/>
    //
    if isfile(tmp)
        if sciVer < 610 // 6.1.0
            doc = xmlRead(tmp);
            setPreferencesValue("/gui2bitmap", ..
                     [attribute ; value], doc);
            xmlWrite(doc);
            xmlDelete(doc);
        else
            xmlSetValues("/gui2bitmap", [attribute; value], tmp)
        end
    else
        t = "<?xml version=""1.0"" encoding=""UTF-8""?>"
        if attribute=="borders" then
            t = [t ; "<gui2bitmap borders="""+value+""" scale=""100""/>"]
        else    // scale
            t = [t ; "<gui2bitmap borders=""[%T %T %T]"" scale="""+value+"""/>"]
        end
        mputl(t, tmp);
    end
endfunction
